@@ -9,6 +9,8 @@
 #include <uuid/uuid.h>
 
 #include "util.h"
+#include "csum.h"
+#include "threads.h"
 
 struct filerec;
 struct block_csum;
@@ -17,7 +19,7 @@ struct results_tree;
 struct dbfile_config;
 
 #define DB_FILE_MAJOR	4
-#define DB_FILE_MINOR	0
+#define DB_FILE_MINOR	1
 
 struct stmts {
 	sqlite3_stmt *insert_block;
@@ -45,6 +47,7 @@ struct stmts {
 	sqlite3_stmt *count_files;
 	sqlite3_stmt *get_max_dedupe_seq;
 	sqlite3_stmt *delete_unscanned_files;
+	sqlite3_stmt *rename_file;
 };
 
 struct dbhandle {
@@ -52,9 +55,28 @@ struct dbhandle {
 	struct stmts stmts;
 };
 
+struct file {
+	int64_t		id;
+	char		filename[PATH_MAX + 1];
+	uint64_t	ino;
+	uint64_t	subvol;
+	size_t		size;
+	uint64_t	blocks;
+	uint64_t	mtime;
+	unsigned int	dedupe_seq;
+	char		digest[DIGEST_LEN];
+	unsigned int	flags;
+};
+
 struct dbhandle *dbfile_open_handle(char *filename);
 void dbfile_close_handle(struct dbhandle *db);
 
+/*
+ * Open the database
+ * On success, the handle is registered to be freed when the thread pool is freed
+ */
+struct dbhandle *dbfile_open_handle_thread(char *filename, struct threads_pool *pool);
+
 void dbfile_lock();
 void dbfile_unlock();
 
@@ -81,15 +103,15 @@ uint64_t count_file_by_digest(struct dbhandle *db, unsigned char *digest,
 				bool show_block_hashes);
 
 struct hash_tree;
-struct hash_file_header;
-struct rb_root;
 
 /*
  * Load hashes into hash_tree only if they have a duplicate in the db.
  * The extent search is later run on the resulting hash_tree.
  */
-int dbfile_load_block_hashes(struct hash_tree *hash_tree, unsigned int seq);
-int dbfile_load_extent_hashes(struct results_tree *res, unsigned int seq);
+int dbfile_load_block_hashes(struct dbhandle *db, struct hash_tree *hash_tree,
+			     unsigned int seq);
+int dbfile_load_extent_hashes(struct dbhandle *db, struct results_tree *res,
+			      unsigned int seq);
 
 struct file_extent {
 	uint64_t	poff;
@@ -102,7 +124,7 @@ int dbfile_load_nondupe_file_extents(struct dbhandle *db, struct filerec *file,
 int dbfile_load_one_file_extent(struct dbhandle *db, struct filerec *file,
 				uint64_t loff, struct file_extent *extent);
 
-int dbfile_load_one_filerec(struct dbhandle *db, uint64_t ino, uint64_t subvol,
+int dbfile_load_one_filerec(struct dbhandle *db, int64_t fileid,
 				struct filerec **file);
 
 /*
@@ -111,8 +133,7 @@ int dbfile_load_one_filerec(struct dbhandle *db, uint64_t ino, uint64_t subvol,
  */
 struct dbhandle *dbfile_get_handle(void);
 
-uint64_t dbfile_store_file_info(struct dbhandle *db, uint64_t ino, uint64_t subvolid,
-			char *path, uint64_t size, uint64_t mtime, unsigned int dedupe_seq);
+int64_t dbfile_store_file_info(struct dbhandle *db, struct file *dbfile);
 int dbfile_store_block_hashes(struct dbhandle *db, int64_t fileid,
 				uint64_t nb_hash, struct block_csum *hashes);
 int dbfile_store_extent_hashes(struct dbhandle *db, int64_t fileid,
@@ -122,7 +143,7 @@ int dbfile_update_scanned_file(struct dbhandle *db, int64_t fileid,
 int dbfile_begin_trans(sqlite3 *db);
 int dbfile_commit_trans(sqlite3 *db);
 int dbfile_abort_trans(sqlite3 *db);
-int dbfile_update_extent_poff(struct dbhandle *db, uint64_t ino, uint64_t subvol,
+int dbfile_update_extent_poff(struct dbhandle *db, int64_t fileid,
 				uint64_t loff, uint64_t poff);
 
 /*
@@ -132,14 +153,17 @@ int dbfile_update_extent_poff(struct dbhandle *db, uint64_t ino, uint64_t subvol
 typedef void (*iter_files_func)(char *filename, char *ino, char *subvol);
 int dbfile_iter_files(struct dbhandle *db, iter_files_func func);
 
-int dbfile_remove_extent_hashes(struct dbhandle *db, uint64_t ino, uint64_t subvolid);
+int dbfile_remove_extent_hashes(struct dbhandle *db, int64_t fileid);
 int dbfile_remove_file(struct dbhandle *db, const char *filename);
 
 void dbfile_list_files(struct dbhandle *db, int (*callback)(void*, int, char**, char**));
 
-int dbfile_describe_file(struct dbhandle *db, uint64_t inum, uint64_t subvolid,
-				uint64_t *mtime, uint64_t *size);
-int dbfile_load_same_files(struct results_tree *res, unsigned int seq);
+int dbfile_describe_file(struct dbhandle *db, uint64_t ino, uint64_t subvol,
+				struct file *dbfile);
+int dbfile_load_same_files(struct dbhandle *db, struct results_tree *res,
+			   unsigned int seq);
+
+int dbfile_rename_file(struct dbhandle *db, int64_t fileid, char *path);
 
 static inline void sqlite3_stmt_cleanup(void *p)
 {
@@ -157,7 +181,7 @@ static inline void sqlite3_reset_stmt(sqlite3_stmt **stmt)
 }
 
 void dbfile_set_gdb(struct dbhandle *db);
-int dbfile_remove_hashes(struct dbhandle *db, uint64_t ino, uint64_t subvolid);
+int dbfile_remove_hashes(struct dbhandle *db, int64_t fileid);
 
 unsigned int get_max_dedupe_seq(struct dbhandle *db);
 int dbfile_prune_unscanned_files(struct dbhandle *db);
