@@ -40,7 +40,7 @@
 #include "dbfile.h"
 #include "memstats.h"
 #include "debug.h"
-
+#include "progress.h"
 #include "file_scan.h"
 #include "find_dupes.h"
 #include "run_dedupe.h"
@@ -74,7 +74,7 @@ static int list_db_files(char *filename)
 
 	_cleanup_(sqlite3_close_cleanup) struct dbhandle *db = dbfile_open_handle(filename);
 	if (!db) {
-		fprintf(stderr, "Error: Could not open \"%s\"\n", filename);
+		eprintf("Error: Could not open \"%s\"\n", filename);
 		return -1;
 	}
 
@@ -97,7 +97,7 @@ static void rm_db_files_from_stdin(struct dbhandle *db)
 		}
 
 		if (readlen > PATH_MAX - 1) {
-			fprintf(stderr, "Path max exceeded: %s\n", path);
+			eprintf("Path max exceeded: %s\n", path);
 			continue;
 		}
 
@@ -110,7 +110,7 @@ static int rm_db_files(int numfiles, char **files)
 	int i, ret;
 	_cleanup_(sqlite3_close_cleanup) struct dbhandle *db = dbfile_open_handle(options.hashfile);
 	if (!db) {
-		fprintf(stderr, "Error: Could not open \"%s\"\n", options.hashfile);
+		eprintf("Error: Could not open \"%s\"\n", options.hashfile);
 		return -1;
 	}
 
@@ -178,7 +178,7 @@ static int parse_dedupe_opts(const char *raw_opts)
 	}
 
 	if (print_usage) {
-		fprintf(stderr, "Bad dedupe options specified. Valid dedupe "
+		eprintf("Bad dedupe options specified. Valid dedupe "
 			"options are:\n"
 			"\t[no]same\n"
 			"\t[no]only_whole_files\n"
@@ -230,7 +230,7 @@ static int process_fdupes()
 		}
 
 		if (readlen > PATH_MAX - 1) {
-			fprintf(stderr, "Path max exceeded: %s\n", path);
+			eprintf("Path max exceeded: %s\n", path);
 			continue;
 		}
 
@@ -255,13 +255,12 @@ static int add_files_from_stdin(struct dbhandle *db)
 		}
 
 		if (readlen > PATH_MAX - 1) {
-			fprintf(stderr, "Path max exceeded: %s\n", path);
+			eprintf("Path max exceeded: %s\n", path);
 			continue;
 		}
 
 		if (scan_file(path, db)) {
-			fprintf(stderr,
-				"Error: cannot add %s into the lookup list\n",
+			eprintf("Error: cannot add %s into the lookup list\n",
 				path);
 			return 1;
 		}
@@ -278,9 +277,7 @@ static int scan_files_from_cmdline(int numfiles, char **files, struct dbhandle *
 		char *name = files[i];
 
 		if (scan_file(name, db)) {
-			fprintf(stderr,
-				"Error: cannot scan %s\n",
-				name);
+			eprintf("Error: cannot scan %s\n", name);
 			return 1;
 		}
 	}
@@ -333,7 +330,7 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 			blocksize = parse_size(optarg);
 			if (blocksize < MIN_BLOCKSIZE ||
 			    blocksize > MAX_BLOCKSIZE){
-				fprintf(stderr, "Error: Blocksize is bounded by %u and %u, %u found\n",
+				eprintf("Error: Blocksize is bounded by %u and %u, %u found\n",
 					MIN_BLOCKSIZE, MAX_BLOCKSIZE, blocksize);
 				return EINVAL;
 			}
@@ -373,7 +370,7 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		case IO_THREADS_OPTION:
 			options.io_threads = strtoul(optarg, NULL, 10);
 			if (!options.io_threads){
-				fprintf(stderr, "Error: --io-threads must be "
+				eprintf("Error: --io-threads must be "
 					"an integer, %s found\n", optarg);
 				return EINVAL;
 			}
@@ -381,7 +378,7 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		case CPU_THREADS_OPTION:
 			options.cpu_threads = strtoul(optarg, NULL, 10);
 			if (!options.cpu_threads){
-				fprintf(stderr, "Error: --cpu-threads must be "
+				eprintf("Error: --cpu-threads must be "
 					"an integer, %s found\n", optarg);
 				return EINVAL;
 			}
@@ -408,7 +405,7 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 			break;
 		case EXCLUDE_OPTION:
 			if (add_exclude_pattern(optarg))
-				fprintf(stderr, "Error: cannot exclude %s\n", optarg);
+				eprintf("Error: cannot exclude %s\n", optarg);
 			break;
 		case BATCH_SIZE_OPTION:
 		case 'B':
@@ -426,14 +423,14 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 	numfiles = argc - optind;
 
 	if (options.only_whole_files && options.do_block_hash) {
-		fprintf(stderr, "Error: using both only_whole_files and partial "
+		eprintf("Error: using both only_whole_files and partial "
 			"options have no meaning\n");
 		return 1;
 	}
 
 	/* Filter out option combinations that don't make sense. */
 	if ((write_hashes + read_hashes + update_hashes) > 1) {
-		fprintf(stderr, "Error: Specify only one hashfile option.\n");
+		eprintf("Error: Specify only one hashfile option.\n");
 		return 1;
 	}
 
@@ -444,10 +441,23 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 	else if (update_hashes)
 		use_hashfile = H_UPDATE;
 
+	/*
+	 * Always add the hashfile and its wal etc to the exclude list
+	 * A wildcard would be easier but may exclude extra files silently,
+	 * this would be confusing for the user.
+	 */
+	if (options.hashfile != NULL) {
+		char tmp[PATH_MAX + 10 ] = {0,};
+		add_exclude_pattern(options.hashfile);
+		snprintf(tmp, PATH_MAX + 9, "%s-wal", options.hashfile);
+		add_exclude_pattern(tmp);
+		snprintf(tmp, PATH_MAX + 9, "%s-shm", options.hashfile);
+		add_exclude_pattern(tmp);
+	}
+
 	if (read_hashes) {
 		if (numfiles) {
-			fprintf(stderr,
-				"Error: --read-hashes option does not take a "
+			eprintf("Error: --read-hashes option does not take a "
 				"file list argument\n");
 			return 1;
 		}
@@ -456,15 +466,13 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 
 	if (options.fdupes_mode) {
 		if (read_hashes || write_hashes || update_hashes) {
-			fprintf(stderr,
-				"Error: cannot mix hashfile option with "
+			eprintf("Error: cannot mix hashfile option with "
 				"--fdupes option\n");
 			return 1;
 		}
 
 		if (numfiles) {
-			fprintf(stderr,
-				"Error: fdupes option does not take a file "
+			eprintf("Error: fdupes option does not take a file "
 				"list argument\n");
 			return 1;
 		}
@@ -477,19 +485,19 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 		stdin_filelist = 1;
 
 	if (list_only_opt && rm_only_opt) {
-		fprintf(stderr, "Error: Can not mix '-L' and '-R' options.\n");
+		eprintf("Error: Can not mix '-L' and '-R' options.\n");
 		return 1;
 	}
 
 	if (list_only_opt || rm_only_opt) {
 		if (!options.hashfile || use_hashfile == H_WRITE) {
-			fprintf(stderr,	"Error: --hashfile= option is required "
+			eprintf("Error: --hashfile= option is required "
 				"with '-L' or -R.\n");
 			return 1;
 		}
 
 		if (list_only_opt && numfiles) {
-			fprintf(stderr, "Error: -L option do not take "
+			eprintf("Error: -L option do not take "
 				"a file list argument\n");
 			return 1;
 		}
@@ -497,7 +505,7 @@ static int parse_options(int argc, char **argv, int *filelist_idx)
 
 	if (!(options.fdupes_mode || list_only_opt)
 			&& numfiles == 0) {
-		fprintf(stderr, "Error: a file list argument is required.\n");
+		eprintf("Error: a file list argument is required.\n");
 		return 1;
 	}
 
@@ -515,7 +523,7 @@ static void print_header(void)
 	qprintf("Gathering file list...\n");
 }
 
-static void __process_duplicates(unsigned int seq)
+static void __process_duplicates(struct dbhandle *db, unsigned int seq)
 {
 	int ret;
 	struct results_tree res;
@@ -525,7 +533,7 @@ static void __process_duplicates(unsigned int seq)
 	init_hash_tree(&dups_tree);
 
 	qprintf("Loading only identical files from hashfile.\n");
-	ret = dbfile_load_same_files(&res, seq + 1);
+	ret = dbfile_load_same_files(db, &res, seq + 1);
 	if (ret)
 		goto out;
 
@@ -542,13 +550,13 @@ static void __process_duplicates(unsigned int seq)
 
 		qprintf("Loading only duplicated hashes from hashfile.\n");
 
-		ret = dbfile_load_extent_hashes(&res, seq + 1);
+		ret = dbfile_load_extent_hashes(db, &res, seq + 1);
 		if (ret)
 			goto out;
 
 		printf("Found %llu identical extents.\n", res.num_extents);
 		if (options.do_block_hash) {
-			ret = dbfile_load_block_hashes(&dups_tree, seq + 1);
+			ret = dbfile_load_block_hashes(db, &dups_tree, seq + 1);
 			if (ret)
 				goto out;
 
@@ -568,16 +576,20 @@ out:
 	free_hash_tree(&dups_tree);
 }
 
-static void process_duplicates()
+static void process_duplicates(struct dbhandle *db)
 {
-	unsigned int max = get_max_dedupe_seq(dbfile_get_handle());
+	unsigned int max = get_max_dedupe_seq(db);
+
+	/* Spawn a dedicated thread pool to block-based lookup */
+	if (options.do_block_hash)
+		extents_search_init();
 
 	for (unsigned int i = dedupe_seq; i < max; i++) {
 		/* Drop all filerecs from the previous iteration. Needed filerecs will be
 		 * recreated by __process_duplicates()
 		 */
 		free_all_filerecs();
-		__process_duplicates(i);
+		__process_duplicates(db, i);
 
 		if (options.run_dedupe) {
 			/*
@@ -587,17 +599,21 @@ static void process_duplicates()
 			dedupe_seq++;
 			dbfile_cfg.dedupe_seq = dedupe_seq;
 			dbfile_cfg.blocksize = blocksize;
-			dbfile_sync_config(dbfile_get_handle(), &dbfile_cfg);
+			dbfile_sync_config(db, &dbfile_cfg);
 		}
 	}
+
+	if (options.do_block_hash)
+		extents_search_free();
 }
 
 static int scan_files(int argc, char **argv, int filelist_idx, struct dbhandle *db)
 {
 	int ret;
 
-	extern bool scan_files_completed;
-	filescan_prepare_pool();
+	filescan_init();
+	if (!quiet)
+		pscan_run();
 
 	if (stdin_filelist)
 		ret = add_files_from_stdin(db);
@@ -605,9 +621,10 @@ static int scan_files(int argc, char **argv, int filelist_idx, struct dbhandle *
 		ret = scan_files_from_cmdline(argc - filelist_idx,
 					     &argv[filelist_idx], db);
 
-	scan_files_completed = true; /* Used to modify the output and print percentages */
-
-	filescan_free_pool();
+	pscan_finish_listing();
+	filescan_free();
+	if (!quiet)
+		pscan_join();
 
 	if (ret)
 		return ret;
@@ -675,7 +692,7 @@ int main(int argc, char **argv)
 	if (use_hashfile == H_WRITE || use_hashfile == H_UPDATE) {
 		ret = dbfile_prune_unscanned_files(db);
 		if (ret) {
-			fprintf(stderr, "Unable to prune unscanned files\n");
+			eprintf("Unable to prune unscanned files\n");
 			goto out;
 		}
 
@@ -688,7 +705,7 @@ int main(int argc, char **argv)
 	}
 
 	if (use_hashfile == H_READ || use_hashfile == H_UPDATE)
-		process_duplicates();
+		process_duplicates(db);
 
 out:
 	free_all_filerecs();
