@@ -41,11 +41,11 @@
 #include <libmount/libmount.h>
 #include <sys/sysmacros.h>
 #include <uuid/uuid.h>
+#include <bsd/sys/queue.h>
 
 #include <glib.h>
 
 #include "csum.h"
-#include "list.h"
 #include "filerec.h"
 #include "hash-tree.h"
 #include "btrfs-util.h"
@@ -56,23 +56,26 @@
 #include "opt.h"
 #include "threads.h"
 #include "fiemap.h"
+#include "progress.h"
 
 /* This is not in linux/magic.h */
 #ifndef	XFS_SB_MAGIC
 #define	XFS_SB_MAGIC		0x58465342	/* 'XFSB' */
 #endif
 
-static int __scan_file(char *path, struct dbhandle *db, struct stat *st);
+struct exclude_file {
+	char *pattern;
+	SLIST_ENTRY(exclude_file) list;
+};
 
-bool scan_files_completed = false;
-static unsigned long long total_files_count = 0;
+SLIST_HEAD(exclude_list, exclude_file) exclude_head = SLIST_HEAD_INITIALIZER(exclude_head);
+
+static int __scan_file(char *path, struct dbhandle *db, struct statx *st);
 
 static struct threads_pool scan_pool;
 
 #define READ_BUF_LEN (8*1024*1024) // 8MB
 
-LIST_HEAD(exclude_list);
-
 struct buffer {
 	char *buf;
 	size_t size; /* Size of buf */
@@ -169,7 +172,7 @@ static int prepare_buffer(struct buffer *buffer)
 	return 0;
 
 err:
-	fprintf(stderr, "prepare_buffer failed\n");
+	eprintf("prepare_buffer failed\n");
 	return 1;
 }
 
@@ -191,29 +194,12 @@ static void free_scan_ctxt(struct scan_ctxt *ctxt)
 		finish_running_checksum(ctxt->extent_csum, NULL);
 }
 
-static void cleanup_dbhandle(void *db)
-{
-	dbfile_close_handle(db);
-}
-
-static struct dbhandle *get_db()
-{
-	struct dbhandle *db;
-
-	dbfile_lock();
-	db = dbfile_open_handle(options.hashfile);
-	dbfile_unlock();
-	if (db)
-		register_cleanup(&scan_pool, (void*)&cleanup_dbhandle, db);
-	return db;
-}
-
 static int is_excluded(const char *name)
 {
-	struct exclude_file *exclude, *tmp;
+	struct exclude_file *exclude;
 
-	list_for_each_entry_safe(exclude, tmp, &exclude_list, list) {
-		if (fnmatch(exclude->pattern, name, FNM_PATHNAME) == 0) {
+	SLIST_FOREACH(exclude, &exclude_head, list) {
+		if (fnmatch(exclude->pattern, name, 0) == 0) {
 			vprintf("Excluding: %s (matches %s)\n", name,
 				exclude->pattern);
 			return 1;
@@ -229,10 +215,15 @@ static inline void mnt_unref_table_cleanup(struct libmnt_table **tb)
 		mnt_unref_table(*tb);
 }
 
+static inline dev_t stx_to_dev(struct statx *stx)
+{
+	return makedev(stx->stx_dev_major, stx->stx_dev_minor);
+}
+
 /* Get the UUID associated with the FS that stores path */
 int get_uuid(char *path, uuid_t *uuid)
 {
-	struct stat st;
+	struct statx st;
 	int ret;
 	_cleanup_(mnt_unref_table_cleanup) struct libmnt_table *tb = NULL;
 	_cleanup_(closefd) int fd = open(path, O_RDONLY);
@@ -241,7 +232,7 @@ int get_uuid(char *path, uuid_t *uuid)
 	struct libmnt_fs *dev = NULL;
 
 	if (fd == -1) {
-		fprintf(stderr, "Cannot open %s: %s\n", path, strerror(errno));
+		eprintf("Cannot open %s: %s\n", path, strerror(errno));
 		return 1;
 	}
 
@@ -249,21 +240,21 @@ int get_uuid(char *path, uuid_t *uuid)
 		dprintf("get_uuid: %s lives on btrfs\n", path);
 		ret = btrfs_get_fsuuid(fd, uuid);
 		if (ret) {
-			fprintf(stderr, "%s: btrfs_get_fsuuid failed\n",
+			eprintf("%s: btrfs_get_fsuuid failed\n",
 				path);
 			return 1;
 		}
 	} else {
 		dprintf("get_uuid: %s do not live on btrfs\n", path);
 
-		ret = lstat(path, &st);
+		ret = statx(0, path, 0, STATX_BASIC_STATS, &st);
 		if (ret) {
-			fprintf(stderr, "Failed to stat %s: %s\n",
+			eprintf("Failed to stat %s: %s\n",
 					path, strerror(errno));
 			return 1;
 		}
 
-		if (major(st.st_dev) == 0) {
+		if (st.stx_dev_major == 0) {
 			dprintf("%s lives on an unsupported filesystem, skipping. "
 				"Please fill a bug if you think this is a mistake.\n",
 					path);
@@ -276,16 +267,16 @@ int get_uuid(char *path, uuid_t *uuid)
 			return 1;
 		}
 
-		dev = mnt_table_find_devno(tb, st.st_dev, MNT_ITER_FORWARD);
+		dev = mnt_table_find_devno(tb, stx_to_dev(&st), MNT_ITER_FORWARD);
 		if (!dev) {
-			fprintf(stderr, "%s: unable to find the mount infos\n",
+			eprintf("%s: unable to find the mount infos\n",
 					path);
 			return 1;
 		}
 
 		uuid_found = blkid_get_tag_value(NULL, "UUID", mnt_fs_get_source(dev));
 		if (!uuid_found) {
-			fprintf(stderr, "libblkid could not get uuid for "
+			eprintf("libblkid could not get uuid for "
 					"device %s. Run blkid as root to "
 					"populate the cache.\n",
 					mnt_fs_get_source(dev));
@@ -297,6 +288,11 @@ int get_uuid(char *path, uuid_t *uuid)
 	return 0;
 }
 
+static inline uint64_t timestamp_to_nano(struct statx_timestamp t)
+{
+	return t.tv_sec * 1000000000 + t.tv_nsec;
+}
+
 /*
  * Check if path lives on a filesystem that is supported, eg
  * that is known to support deduplication.
@@ -308,7 +304,7 @@ bool is_fs_supported(char *path)
 
 	ret = statfs(path, &fs);
 	if (ret) {
-		fprintf(stderr, "Error %d: %s while check fs type on %s",
+		eprintf("Error %d: %s while check fs type on %s",
 			errno, strerror(errno), path);
 		return false;
 	}
@@ -326,7 +322,7 @@ bool is_fs_supported(char *path)
  *
  * Returns true is the file is legit, false if not (or on error)
  */
-bool check_file(struct dbhandle *db, char *path, struct stat *st, bool parent_checked)
+bool check_file(struct dbhandle *db, char *path, struct statx *st, bool parent_checked)
 {
 	int ret;
 	struct dbfile_config cfg;
@@ -335,12 +331,12 @@ bool check_file(struct dbhandle *db, char *path, struct stat *st, bool parent_ch
 	if (is_excluded(path))
 		return false;
 
-	if (!S_ISREG(st->st_mode) && !S_ISDIR(st->st_mode)) {
+	if (!S_ISREG(st->stx_mode) && !S_ISDIR(st->stx_mode)) {
 		vprintf("Skipping non-regular/non-directory file %s\n", path);
 		return false;
 	}
 
-	if (S_ISREG(st->st_mode) && st->st_size == 0) {
+	if (S_ISREG(st->stx_mode) && st->stx_size == 0) {
 		vprintf("Skipping empty file %s\n", path);
 		return false;
 	}
@@ -348,7 +344,7 @@ bool check_file(struct dbhandle *db, char *path, struct stat *st, bool parent_ch
 	/* There is no need to check if the file lives in our locked fs.
 	 * It is a regular file and we already check its parent.
 	 */
-	if (S_ISREG(st->st_mode) && parent_checked)
+	if (S_ISREG(st->stx_mode) && parent_checked)
 		return true;
 
 	/* Locked-fs checks */
@@ -370,11 +366,11 @@ bool check_file(struct dbhandle *db, char *path, struct stat *st, bool parent_ch
 		if (ret)
 			return false;
 
-		locked_fs.dev = st->st_dev;
+		locked_fs.dev = stx_to_dev(st);
 		locked_fs.is_btrfs = is_btrfs(path);
 
 		if (!is_fs_supported(path))
-			fprintf(stderr, "Warn: filesystem for %s is not known to "
+			eprintf("Warn: filesystem for %s is not known to "
 				"support deduplication.\n", path);
 
 		return true;
@@ -390,21 +386,21 @@ bool check_file(struct dbhandle *db, char *path, struct stat *st, bool parent_ch
 			return false;
 
 		if (uuid_compare(uuid, locked_fs.uuid) != 0) {
-			fprintf(stderr, "%s lives on fs ", path);
-			debug_print_uuid(stderr, uuid);
-			fprintf(stderr, " will we are locked on fs ");
-			debug_print_uuid(stderr, locked_fs.uuid);
-			fprintf(stderr, ".\n");
+			eprintf("%s lives on fs ", path);
+			debug_print_uuid(uuid);
+			eprintf(" will we are locked on fs ");
+			debug_print_uuid(locked_fs.uuid);
+			eprintf(".\n");
 			return false;
 		}
 
-		locked_fs.dev = st->st_dev;
+		locked_fs.dev = stx_to_dev(st);
 		locked_fs.is_btrfs = is_btrfs(path);
 		return true;
 	}
 
 	if (!locked_fs.is_btrfs)
-		return locked_fs.dev == st->st_dev;
+		return locked_fs.dev == stx_to_dev(st);
 
 	/* On btrfs, we must always fetch the UUID */
 	ret = get_uuid(path, &uuid);
@@ -423,7 +419,7 @@ void fs_get_locked_uuid(uuid_t *uuid)
 static int get_dirent_type(struct dirent *entry, int fd, const char *path)
 {
 	int ret;
-	struct stat st;
+	struct statx st;
 
 	if (entry->d_type != DT_UNKNOWN)
 		return entry->d_type;
@@ -433,28 +429,27 @@ static int get_dirent_type(struct dirent *entry, int fd, const char *path)
 	 * fashioned way. We translate mode to DT_* for the
 	 * convenience of the caller.
 	 */
-	ret = fstatat(fd, entry->d_name, &st, AT_SYMLINK_NOFOLLOW);
-	if (ret) {
-		fprintf(stderr,
-			"Error %d: %s while getting type of file %s/%s. "
+	ret = statx(fd, entry->d_name, AT_SYMLINK_NOFOLLOW, STATX_BASIC_STATS, &st);
+	if (ret || !(st.stx_mask & STATX_BASIC_STATS)) {
+		eprintf("Error %d: %s while getting type of file %s/%s. "
 			"Skipping.\n",
 			errno, strerror(errno), path, entry->d_name);
 		return DT_UNKNOWN;
 	}
 
-	if (S_ISREG(st.st_mode))
+	if (S_ISREG(st.stx_mode))
 		return DT_REG;
-	if (S_ISDIR(st.st_mode))
+	if (S_ISDIR(st.stx_mode))
 		return DT_DIR;
-	if (S_ISBLK(st.st_mode))
+	if (S_ISBLK(st.stx_mode))
 		return DT_BLK;
-	if (S_ISCHR(st.st_mode))
+	if (S_ISCHR(st.stx_mode))
 		return DT_CHR;
-	if (S_ISFIFO(st.st_mode))
+	if (S_ISFIFO(st.stx_mode))
 		return DT_FIFO;
-	if (S_ISLNK(st.st_mode))
+	if (S_ISLNK(st.stx_mode))
 		return DT_LNK;
-	if (S_ISSOCK(st.st_mode))
+	if (S_ISSOCK(st.stx_mode))
 		return DT_SOCK;
 
 	return DT_UNKNOWN;
@@ -467,14 +462,14 @@ static int walk_dir(char *path, struct dbhandle *db)
 {
 	int ret = 0;
 	struct dirent *entry;
-	struct stat st;
+	struct statx st;
 	_cleanup_(closedirectory) DIR *dirp = opendir(path);
 
 	/* Overallocate to peace the compiler. An abort will check the actual values. */
 	char child[PATH_MAX + 257] = { 0, };
 
 	if (dirp == NULL) {
-		fprintf(stderr, "Error %d: %s while opening directory %s\n",
+		eprintf("Error %d: %s while opening directory %s\n",
 			errno, strerror(errno), path);
 		return 0;
 	}
@@ -486,7 +481,7 @@ static int walk_dir(char *path, struct dbhandle *db)
 			break;
 
 		if (errno != 0) {
-			fprintf(stderr, "Error %d: %s while reading directory %s\n",
+			eprintf("Error %d: %s while reading directory %s\n",
 				errno, strerror(errno), path);
 			return 0;
 		}
@@ -509,9 +504,9 @@ static int walk_dir(char *path, struct dbhandle *db)
 		else
 			sprintf(child, "%s/%s", path, entry->d_name);
 
-		ret = lstat(child, &st);
-		if (ret) {
-			fprintf(stderr, "Failed to stat %s: %s\n",
+		ret = statx(0, child, 0, STATX_BASIC_STATS, &st);
+		if (ret || !(st.stx_mask | STATX_BASIC_STATS)) {
+			eprintf("Failed to stat %s: %s\n",
 					path, strerror(errno));
 			continue;
 		}
@@ -530,22 +525,36 @@ static int walk_dir(char *path, struct dbhandle *db)
 	return 0;
 }
 
+static inline bool is_file_renamed(char *path_in_db, char *path)
+{
+	struct stat st;
+
+	if (strlen(path_in_db) == 0 || strcmp(path_in_db, path) == 0)
+		return false;
+
+	/*
+	 * Old path and new paths differs. Could be hardlink,
+	 * so we check if the old still exists.
+	 */
+	return true ? lstat(path_in_db, &st) : false;
+}
+
 /*
  * Returns nonzero on fatal errors only
  * This function schedules csum_whole_file()
  * The caller must call check_file() before and must not call
  * this if path is not a regular file.
  */
-static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
+static int __scan_file(char *path, struct dbhandle *db, struct statx *st)
 {
 	int ret;
-	uint64_t mtime = 0, size = 0;
+	struct file dbfile = {0,};
 	static unsigned int seq = 0, counter = 0;
 	GError *err = NULL;
 	struct file_to_scan *file;
 	int64_t fileid = 0;
-
-	uint64_t subvolid = 0;
+	bool file_renamed;
+	static uint64_t position = 0;
 
 	/*
 	 * The first call initializes the static variable
@@ -555,13 +564,13 @@ static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
 	if (seq == 0)
 		seq = dedupe_seq + 1;
 
-	abort_on(!S_ISREG(st->st_mode));
+	abort_on(!S_ISREG(st->stx_mode));
 
 	if (locked_fs.is_btrfs) {
 		_cleanup_(closefd) int fd;
 		fd = open(path, O_RDONLY);
 		if (fd == -1) {
-			fprintf(stderr, "Error %d: %s while opening file \"%s\". "
+			eprintf("Error %d: %s while opening file \"%s\". "
 				"Skipping.\n", errno, strerror(errno), path);
 			return 0;
 		}
@@ -571,10 +580,9 @@ static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
 		 * can have the same i_ino. Get the subvolume id of
 		 * our file so hard link detection works.
 		 */
-		ret = lookup_btrfs_subvolid(fd, &subvolid);
+		ret = lookup_btrfs_subvol(fd, &(dbfile.subvol));
 		if (ret) {
-			fprintf(stderr,
-				"Error %d: %s while finding subvolid for file "
+			eprintf("Error %d: %s while finding subvol for file "
 				"\"%s\". Skipping.\n", ret, strerror(ret),
 				path);
 			return 0;
@@ -584,14 +592,17 @@ static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
 	/*
 	 * Check the database to see if that file need rescan or not.
 	 */
-	ret = dbfile_describe_file(db, st->st_ino, subvolid, &mtime, &size);
+	ret = dbfile_describe_file(db, st->stx_ino, dbfile.subvol, &dbfile);
 	if (ret) {
 		vprintf("dbfile_describe_file failed\n");
 		return 0;
 	}
 
+	file_renamed = is_file_renamed(dbfile.filename, path);
+
 	/* Database is up-to-date, nothing more to do */
-	if (mtime == timespec_to_nano(&(st->st_mtim)) && size == (uint64_t)st->st_size)
+	if (dbfile.mtime == timestamp_to_nano(st->stx_mtime)
+	    && dbfile.size == st->stx_size && !file_renamed)
 		return 0;
 
 	if (options.batch_size != 0) {
@@ -602,18 +613,33 @@ static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
 		}
 	}
 
+	dbfile.ino = st->stx_ino;
+	dbfile.size = st->stx_size;
+	strncpy(dbfile.filename, path, PATH_MAX);
+	dbfile.mtime = timestamp_to_nano(st->stx_mtime);
+	dbfile.dedupe_seq = seq;
+
 	dbfile_lock();
 	dbfile_begin_trans(db->db);
-	if (mtime != 0 || size != 0) {
+
+	if (file_renamed) {
+		ret = dbfile_rename_file(db, dbfile.id, path);
+		if (ret) {
+			vprintf("dbfile_rename_file failed\n");
+			return 0;
+		}
+	}
+
+	if (dbfile.mtime != 0 || dbfile.size != 0) {
 		/*
 		 * The file was scanned in a previous run.
 		 * We will rescan it, so let's remove old hashes
 		 */
-		dbfile_remove_hashes(db, st->st_ino, subvolid);
+		dbfile_remove_hashes(db, dbfile.id);
 	}
 
 	/* Upsert the file record */
-	fileid = dbfile_store_file_info(db, st->st_ino, subvolid, path, st->st_size, timespec_to_nano(&(st->st_mtim)), seq);
+	fileid = dbfile_store_file_info(db, &dbfile);
 	if (!fileid) {
 		dbfile_abort_trans(db->db);
 		dbfile_unlock();
@@ -628,16 +654,18 @@ static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
 
 	file->path = strdup(path);
 	file->fileid = fileid;
-	file->filesize = st->st_size;
+	file->filesize = st->stx_size;
 
-	total_files_count++;
-	file->file_position = total_files_count;
+	pscan_set_progress(1, st->stx_size);
+	position++;
+	file->file_position = position;
 
 	if(!g_thread_pool_push(scan_pool.pool, file, &err)) {
-		fprintf(stderr, "g_thread_pool_push: %s\n", err->message);
+		eprintf("g_thread_pool_push: %s\n", err->message);
 		g_error_free(err);
 		err = NULL;
 		free(file);
+		return 1;
 	}
 
 	return 0;
@@ -646,7 +674,7 @@ static int __scan_file(char *path, struct dbhandle *db, struct stat *st)
 /* The entry point for files passed by the user */
 int scan_file(char *in_path, struct dbhandle *db)
 {
-	struct stat st;
+	struct statx st;
 	char path[PATH_MAX];
 	int ret;
 
@@ -660,15 +688,15 @@ int scan_file(char *in_path, struct dbhandle *db)
 	 *   any directory.
 	 */
 	if (realpath(in_path, path) == NULL) {
-		fprintf(stderr, "Error %d: %s while getting path to file %s. "
+		eprintf("Error %d: %s while getting path to file %s. "
 			"Skipping.\n",
 			errno, strerror(errno), in_path);
 		return 0;
 	}
 
-	ret = lstat(path, &st);
-	if (ret) {
-		fprintf(stderr, "Error %d: %s while stating file %s. "
+	ret = statx(0, path, 0, STATX_BASIC_STATS, &st);
+	if (ret || !(st.stx_mask & STATX_BASIC_STATS)) {
+		eprintf("Error %d: %s while stating file %s. "
 			"Skipping.\n",
 			errno, strerror(errno), path);
 		return 0;
@@ -677,7 +705,7 @@ int scan_file(char *in_path, struct dbhandle *db)
 	if (!check_file(db, path, &st, false))
 		return 0;
 
-	if (S_ISREG(st.st_mode))
+	if (S_ISREG(st.stx_mode))
 		return __scan_file(path, db, &st);
 	else
 		return walk_dir(path, db);
@@ -709,26 +737,6 @@ static int add_block_hash(struct hashes *hashes,
 	return 0;
 }
 
-static void print_progress(unsigned long long pos, char* path)
-{
-	unsigned int leading_space = num_digits(total_files_count);
-
-	/*
-	 * We do not print the percentage unless all files are actually
-	 * pushed into the thread queue
-	 */
-	if (scan_files_completed) {
-		qprintf("[%0*llu/%llu] (%05.2f%%) csum: %s\n",
-			leading_space, pos, total_files_count,
-			(double)pos / (double)total_files_count * 100,
-			path);
-	} else {
-		qprintf("[%0*llu/%llu] csum: %s\n",
-			leading_space, pos, total_files_count,
-			path);
-	}
-}
-
 /*
  * Check if the area should be scanned.
  */
@@ -843,7 +851,7 @@ static int process_extents(struct scan_ctxt *ctxt, struct buffer *buffer,
 	while (file_off < ctxt->off + bytes) {
 		extent = get_extent(ctxt->fiemap, file_off, NULL);
 		if (!extent) {
-			fprintf(stderr, "process_extents: unable to get extent\n");
+			eprintf("process_extents: unable to get extent\n");
 
 			/* Cleanup the partial checksum and skip
 			 * the rest of the buffer
@@ -967,8 +975,10 @@ static void csum_whole_file(struct file_to_scan *file)
 	 */
 	static struct dbhandle *db = NULL;
 	static __thread struct buffer buffer = {0,};
+	static __thread struct pscan_thread *tls_progress = NULL;
 
 	/* Dummy variables used to trigger the cleanup code */
+	_cleanup_(pscan_reset_thread) struct pscan_thread *tprogress = tls_progress;
 	_cleanup_(freep) char *path = file->path;
 	_cleanup_(freep) struct file_to_scan *clean_file = file;
 
@@ -977,12 +987,13 @@ static void csum_whole_file(struct file_to_scan *file)
 	 */
 	bool eof_reached = false;
 
-	print_progress(file->file_position, file->path);
+	/* Prevent close on fd 0 if, somehow, an error occurs before we open */
+	ctxt.fd = -1;
 
 	if (!(buffer.buf)) {
 		ret = prepare_buffer(&buffer);
 		if (ret) {
-			fprintf(stderr, "unable to prepare our read buffer\n");
+			eprintf("unable to prepare our read buffer\n");
 			return;
 		}
 	} else {
@@ -992,12 +1003,23 @@ static void csum_whole_file(struct file_to_scan *file)
 	}
 
 	if (!db)
-		db = get_db();
+		db = dbfile_open_handle_thread(options.hashfile, &scan_pool);
 	if (!db) {
-		fprintf(stderr, "csum_whole_file: unable to connect to the database\n");
+		eprintf("csum_whole_file: unable to connect to the database\n");
 		return;
 	}
 
+	if (!tls_progress) {
+		tls_progress = pscan_register_thread(gettid());
+		abort_on(!tls_progress);
+		tprogress = tls_progress;
+	}
+
+	tprogress->status = thread_scanning;
+	tprogress->file_scanned_bytes = 0;
+	tprogress->file_total_bytes = file->filesize;
+	strncpy(tprogress->file_path, file->path, PATH_MAX);
+
 	ctxt.filesize = file->filesize;
 	ctxt.file_csum = start_running_checksum();
 	if (!ctxt.file_csum)
@@ -1005,7 +1027,7 @@ static void csum_whole_file(struct file_to_scan *file)
 
 	ctxt.fd = open(file->path, O_RDONLY);
 	if (ctxt.fd == -1) {
-		fprintf(stderr, "csum_whole_file: Error %d: %s while opening file \"%s\". "
+		eprintf("csum_whole_file: Error %d: %s while opening file \"%s\". "
 			"Skipping.\n", errno, strerror(errno), file->path);
 		return;
 	}
@@ -1015,7 +1037,7 @@ static void csum_whole_file(struct file_to_scan *file)
 		return;
 
 	if (!allocate_hashes(&hashes, &ctxt)) {
-		fprintf(stderr, "allocate_hashes failed\n");
+		eprintf("allocate_hashes failed\n");
 		return;
 	}
 
@@ -1038,7 +1060,7 @@ static void csum_whole_file(struct file_to_scan *file)
 		ret = fill_buffer(&ctxt, &buffer);
 		if (ret < 0) {
 			ret = errno;
-			fprintf(stderr, "Unable to read file %s: %s\n",
+			eprintf("Unable to read file %s: %s\n",
 				file->path, strerror(ret));
 			return;
 		}
@@ -1048,10 +1070,13 @@ static void csum_whole_file(struct file_to_scan *file)
 
 		bytes_processed = process_blocks(&ctxt, &buffer, &hashes);
 		if (bytes_processed < 0) {
-			fprintf(stderr, "process_blocks failed somehow\n");
+			eprintf("process_blocks failed somehow\n");
 			return;
 		}
 
+		tprogress->file_scanned_bytes += bytes_processed;
+		tprogress->total_scanned_bytes += bytes_processed;
+
 		/* Process the last partial block */
 		if (eof_reached && (size_t)bytes_processed < buffer.dl_len) {
 			ret = process_block(buffer.buf + bytes_processed,
@@ -1059,7 +1084,7 @@ static void csum_whole_file(struct file_to_scan *file)
 					    ctxt.off + bytes_processed,
 					    &hashes);
 			if (ret) {
-				fprintf(stderr, "Unable to process %s's last block\n", file->path);
+				eprintf("Unable to process %s's last block\n", file->path);
 				return;
 			}
 
@@ -1086,14 +1111,16 @@ static void csum_whole_file(struct file_to_scan *file)
 	}
 
 	if (ctxt.off != ctxt.filesize) {
-		fprintf(stderr, "file %s changed\n", file->path);
+		eprintf("file %s changed\n", file->path);
 		return;
 	}
 
 	finish_running_checksum(ctxt.file_csum, file_digest);
 	ctxt.file_csum = NULL;
 
+	tprogress->status = thread_waiting_lock;
 	dbfile_lock();
+	tprogress->status = thread_committing;
 	ret = dbfile_begin_trans(db->db);
 	if (ret) {
 		dbfile_unlock();
@@ -1159,7 +1186,7 @@ int add_exclude_pattern(const char *pattern)
 		getcwd(cwd, PATH_MAX);
 
 		if (strlen(cwd) + strlen(pattern) > PATH_MAX) {
-			fprintf(stderr, "Error: cannot prepend cwd to %s\n", pattern);
+			eprintf("Error: cannot prepend cwd to %s\n", pattern);
 			return 1;
 		}
 
@@ -1167,27 +1194,33 @@ int add_exclude_pattern(const char *pattern)
 		exclude->pattern = strdup(exp_pattern);
 	}
 
-	printf("Adding exclude pattern: %s\n", exclude->pattern);
+	vprintf("Adding exclude pattern: %s\n", exclude->pattern);
 
-	list_add_tail(&exclude->list, &exclude_list);
+	SLIST_INSERT_HEAD(&exclude_head, exclude, list);
 	return 0;
 }
 
-void filescan_prepare_pool()
+void filescan_init(void)
 {
 	abort_on(scan_pool.pool);
-	setup_pool(&scan_pool, csum_whole_file, NULL);
+	setup_pool(&scan_pool, csum_whole_file, NULL, options.io_threads);
 	abort_on(!scan_pool.pool);
 }
 
-void filescan_free_pool()
+void filescan_free(void)
 {
 	free_pool(&scan_pool);
 }
 
 void add_file_fdupes(char *path)
 {
-	struct stat st;
-	lstat(path, &st);
-	filerec_new(path, st.st_ino, 0, st.st_size);
+	struct statx st;
+	int ret;
+
+	ret = statx(AT_FDCWD, path, 0, STATX_BASIC_STATS, &st);
+	if (ret || !(st.stx_mask & STATX_BASIC_STATS)) {
+		eprintf("statx on %s: %s\n", path, strerror(errno));
+		return;
+	}
+	filerec_new(path, st.stx_ino, st.stx_size);
 }
